from sympy import Integer, Symbol, Function
from lib.symbolic import Dist

relu_function = Function("ReLU")
soft_function = Function("Soft")

class ReluModel:
    def __init__(self, count: int, with_bias: bool = True):
        self.count = count
        self.with_bias = with_bias
    
    def get(self, dist: Dist):
        pre_expr = Integer(0)
        post_expr = Integer(0)
        template_vars = []
        state_vars = []
        bias_vars = []

        for i in range(0, self.count):
            relu_pre_expr = 0
            relu_post_expr = 0
            if self.with_bias:
                const = Symbol(f"b_{i}")
                relu_pre_expr += const
                relu_post_expr += const
                template_vars.append(const)
                bias_vars.append(const)
            for k in dist.begin_map:
                coefficient = Symbol(f"w_{i},{k}")
                relu_pre_expr += coefficient * Symbol(f"{k}_0")
                relu_post_expr += coefficient * Symbol(f"{k}_{dist.vars_map[k]}")
                template_vars.append(coefficient)
                state_vars.append(Symbol(f"{k}_0"))
            pre_expr += relu_function(relu_pre_expr)
            post_expr += relu_function(relu_post_expr)

        return state_vars, template_vars, pre_expr, post_expr, bias_vars

class TwoLayerReluModel:
    def __init__(self, count: int):
        self.count = count
    
    def get(self, dist: Dist):
        pre_expr = Integer(0)
        post_expr = Integer(0)
        template_vars = []
        state_vars = []
        bias_vars = []

        for i in range(0, self.count):
            const = Symbol(f"b_{i}")
            weight = Symbol(f"c_{i}")
            relu_pre_expr = const
            relu_post_expr = const
            template_vars.append(const)
            template_vars.append(weight)
            bias_vars.append(const)
            for k in dist.begin_map:
                coefficient = Symbol(f"w_{i},{k}")
                relu_pre_expr += coefficient * Symbol(f"{k}_0")
                relu_post_expr += coefficient * Symbol(f"{k}_{dist.vars_map[k]}")
                template_vars.append(coefficient)
                state_vars.append(Symbol(f"{k}_0"))
            pre_expr += weight * relu_function(relu_pre_expr)
            post_expr += weight * relu_function(relu_post_expr)

        pre_expr = relu_function(pre_expr)
        post_expr = relu_function(post_expr)
        return state_vars, template_vars, pre_expr, post_expr, bias_vars

class SosModel:
    def __init__(self, count: int):
        self.count = count
    
    def get(self, dist: Dist):
        pre_expr = Integer(0)
        post_expr = Integer(0)
        template_vars = []
        state_vars = []
        bias_vars = []

        for i in range(0, self.count):
            const = Symbol(f"b_{i}")
            relu_pre_expr = const
            relu_post_expr = const
            template_vars.append(const)
            bias_vars.append(const)
            for k in dist.begin_map:
                coefficient = Symbol(f"w_{i},{k}")
                relu_pre_expr += coefficient * Symbol(f"{k}_0")
                relu_post_expr += coefficient * Symbol(f"{k}_{dist.vars_map[k]}")
                template_vars.append(coefficient)
                state_vars.append(Symbol(f"{k}_0"))
            pre_expr += relu_pre_expr ** 2
            post_expr += relu_post_expr ** 2

        return state_vars, template_vars, pre_expr, post_expr, bias_vars